Package: dblog
Title: Logging for DBI
Version: 0.0.0.9013
Date: 2023-10-09
Authors@R: 
    person(given = "Kirill",
           family = "Müller",
           role = c("aut", "cre"),
           email = "krlmlr+r@mailbox.org",
           comment = c(ORCID = "0000-0002-1416-3412"))
Description: Provides logging of DBI methods for arbitrary backends.
License: GPL-3
URL: https://dblog.r-dbi.org, https://github.com/r-dbi/dblog
BugReports: https://github.com/r-dbi/dblog/issues
Imports: 
    collections (>= 0.1.4),
    DBI,
    glue,
    methods,
    purrr,
    rlang
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.2.3
Collate: 
    'Driver.R'
    'Connection.R'
    'Result.R'
    'deparse.R'
    'import.R'
    'make-log-call.R'
Suggests: 
    testthat (>= 3.0.0)
Config/testthat/edition: 3

# Generated by roxygen2: do not edit by hand

S3method(format,dblogConnection)
export(dblog)
export(dblog_cnr)
export(get_default_logger)
export(make_collect_logger)
export(make_text_logger)
import(DBI)
import(methods)
import(rlang)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# dblog

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://www.tidyverse.org/lifecycle/#experimental)
[![CRAN
status](https://www.r-pkg.org/badges/version/dblog)](https://cran.r-project.org/package=dblog)
<!-- badges: end -->

The goal of dblog is to implement logging for arbitrary DBI backends,
similarly to Perl’s [DBI::Log](https://metacpan.org/pod/DBI::Log). This
is useful for troubleshooting and auditing codes that access a database.
The initial use case for this package is to help debugging DBItest
tests.

## Installation

You can install the released version of dblog from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("dblog")
```

Install the development version from GitHub using

``` r
# install.packages("devtools")
devtools::install_github("r-dbi/dblog")
```

## Basic example

The `dblog` driver wraps arbitrary drivers:

``` r
library(dblog)
drv <- dblog(RSQLite::SQLite())
#> drv1 <- RSQLite::SQLite()
```

All calls to DBI methods are logged, by default to the console.

``` r
conn <- dbConnect(drv, file = ":memory:")
#> conn1 <- dbConnect(drv1, file = ":memory:")
dbWriteTable(conn, "iris", iris[1:3, ])
#> dbWriteTable(conn1, name = "iris", value = structure(list(Sepal.Length = c(5.1, 4.9, 
#> 4.7), Sepal.Width = c(3.5, 3, 3.2), Petal.Length = c(1.4, 1.4, 1.3), Petal.Width = c(0.2, 
#> 0.2, 0.2), Species = structure(c(1L, 1L, 1L), .Label = c("setosa", "versicolor", 
#> "virginica"), class = "factor")), row.names = c(NA, 3L), class = "data.frame"), overwrite = FALSE, 
#>     append = FALSE)
data <- dbGetQuery(conn, "SELECT * FROM iris")
#> dbGetQuery(conn1, "SELECT * FROM iris")
#> ##   Sepal.Length Sepal.Width Petal.Length Petal.Width Species
#> ## 1          5.1         3.5          1.4         0.2  setosa
#> ## 2          4.9         3.0          1.4         0.2  setosa
#> ## 3          4.7         3.2          1.3         0.2  setosa
dbDisconnect(conn)
#> dbDisconnect(conn1)

data
#>   Sepal.Length Sepal.Width Petal.Length Petal.Width Species
#> 1          5.1         3.5          1.4         0.2  setosa
#> 2          4.9         3.0          1.4         0.2  setosa
#> 3          4.7         3.2          1.3         0.2  setosa
```

The log is runnable R code\! Run it in a fresh session to repeat the
operations, step by step or in an otherwise controlled fashion.

dblog is smart about DBI objects created or returned, and will assign a
new variable name to each new object. Cleared results or closed
connections are not removed automatically.

## Logging options

Logging can be redirected to a file, optionally all outputs may be
logged as well. For example, use a collecting logger to output all calls
and results after the fact.

``` r
collecting_logger <- make_collect_logger()

drv <- dblog(RSQLite::SQLite(), logger = collecting_logger)
conn <- dbConnect(drv, file = ":memory:")
dbWriteTable(conn, "iris", iris[1:3, ])
data <- dbGetQuery(conn, "SELECT * FROM iris")
dbDisconnect(conn)

collecting_logger$retrieve()
#> drv1 <- RSQLite::SQLite()
#> conn1 <- dbConnect(drv1, file = ":memory:")
#> dbWriteTable(conn1, name = "iris", value = structure(list(Sepal.Length = c(5.1, 4.9, 
#> 4.7), Sepal.Width = c(3.5, 3, 3.2), Petal.Length = c(1.4, 1.4, 1.3), Petal.Width = c(0.2, 
#> 0.2, 0.2), Species = structure(c(1L, 1L, 1L), .Label = c("setosa", "versicolor", 
#> "virginica"), class = "factor")), row.names = c(NA, 3L), class = "data.frame"), overwrite = FALSE, 
#>     append = FALSE)
#> dbGetQuery(conn1, "SELECT * FROM iris")
#> ##   Sepal.Length Sepal.Width Petal.Length Petal.Width Species
#> ## 1          5.1         3.5          1.4         0.2  setosa
#> ## 2          4.9         3.0          1.4         0.2  setosa
#> ## 3          4.7         3.2          1.3         0.2  setosa
#> dbDisconnect(conn1)

ev <- evaluate::evaluate(collecting_logger$retrieve())
cat(unlist(ev, use.names = FALSE), sep = "\n")
#> drv1 <- RSQLite::SQLite()
#> 
#> conn1 <- dbConnect(drv1, file = ":memory:")
#> 
#> dbWriteTable(conn1, name = "iris", value = structure(list(Sepal.Length = c(5.1, 4.9, 
#> 4.7), Sepal.Width = c(3.5, 3, 3.2), Petal.Length = c(1.4, 1.4, 1.3), Petal.Width = c(0.2, 
#> 0.2, 0.2), Species = structure(c(1L, 1L, 1L), .Label = c("setosa", "versicolor", 
#> "virginica"), class = "factor")), row.names = c(NA, 3L), class = "data.frame"), overwrite = FALSE, 
#>     append = FALSE)
#> 
#> dbGetQuery(conn1, "SELECT * FROM iris")
#> 
#>   Sepal.Length Sepal.Width Petal.Length Petal.Width Species
#> 1          5.1         3.5          1.4         0.2  setosa
#> 2          4.9         3.0          1.4         0.2  setosa
#> 3          4.7         3.2          1.3         0.2  setosa
#> 
#> ##   Sepal.Length Sepal.Width Petal.Length Petal.Width Species
#> 
#> ## 1          5.1         3.5          1.4         0.2  setosa
#> 
#> ## 2          4.9         3.0          1.4         0.2  setosa
#> 
#> ## 3          4.7         3.2          1.3         0.2  setosa
#> 
#> dbDisconnect(conn1)
```

## Logging complex operations

The full power is demonstrated when running with code where the
underlying *DBI* operations are not obvious:

``` r
library(dplyr)
#> 
#> Attaching package: 'dplyr'
#> The following objects are masked from 'package:stats':
#> 
#>     filter, lag
#> The following objects are masked from 'package:base':
#> 
#>     intersect, setdiff, setequal, union

drv <- dblog(RSQLite::SQLite())
#> drv2 <- RSQLite::SQLite()
conn <- dbConnect(drv, file = ":memory:")
#> conn1 <- dbConnect(drv2, file = ":memory:")
dbWriteTable(conn, "iris", iris[1:3, ])
#> dbWriteTable(conn1, name = "iris", value = structure(list(Sepal.Length = c(5.1, 4.9, 
#> 4.7), Sepal.Width = c(3.5, 3, 3.2), Petal.Length = c(1.4, 1.4, 1.3), Petal.Width = c(0.2, 
#> 0.2, 0.2), Species = structure(c(1L, 1L, 1L), .Label = c("setosa", "versicolor", 
#> "virginica"), class = "factor")), row.names = c(NA, 3L), class = "data.frame"), overwrite = FALSE, 
#>     append = FALSE)

src <- dbplyr::src_dbi(conn)
iris_tbl <- tbl(src, "iris")
#> res1 <- dbSendQuery(conn1, structure("SELECT *\nFROM `iris` AS `zzz1`\nWHERE (0 = 1)", class = c("sql", 
#> "character")))
#> dbFetch(res1, n = 0)
#> ## [1] Sepal.Length Sepal.Width  Petal.Length Petal.Width  Species     
#> ## <0 rows> (or 0-length row.names)
#> dbClearResult(res1)
iris_tbl %>%
  summarize_if(is.numeric, mean)
#> Applying predicate on the first 100 rows
#> res2 <- dbSendQuery(conn1, structure("SELECT *\nFROM `iris`\nLIMIT 100", class = c("sql", 
#> "character")))
#> dbFetch(res2, n = 100)
#> ##   Sepal.Length Sepal.Width Petal.Length Petal.Width Species
#> ## 1          5.1         3.5          1.4         0.2  setosa
#> ## 2          4.9         3.0          1.4         0.2  setosa
#> ## 3          4.7         3.2          1.3         0.2  setosa
#> dbHasCompleted(res2)
#> ## [1] TRUE
#> dbClearResult(res2)
#> Warning: Missing values are always removed in SQL.
#> Use `mean(x, na.rm = TRUE)` to silence this warning
#> This warning is displayed only once per session.
#> res3 <- dbSendQuery(conn1, structure("SELECT AVG(`Sepal.Length`) AS `Sepal.Length`, AVG(`Sepal.Width`) AS `Sepal.Width`, AVG(`Petal.Length`) AS `Petal.Length`, AVG(`Petal.Width`) AS `Petal.Width`\nFROM `iris`\nLIMIT 11", class = c("sql", 
#> "character")))
#> dbFetch(res3, n = -1)
#> ##   Sepal.Length Sepal.Width Petal.Length Petal.Width
#> ## 1          4.9    3.233333     1.366667         0.2
#> dbHasCompleted(res3)
#> ## [1] TRUE
#> dbClearResult(res3)
#> # Source:   lazy query [?? x 4]
#> # Database: sqlite 3.29.0 []
#>   Sepal.Length Sepal.Width Petal.Length Petal.Width
#>          <dbl>       <dbl>        <dbl>       <dbl>
#> 1         4.90        3.23         1.37         0.2
```

## Inheritance hierarchy

Despite the common suggestion to [prefer composition over
inheritance](https://en.wikipedia.org/wiki/Composition_over_inheritance),
the new logging classes are implemented as subclasses of the actual DBI
classes. Moreover, the class definitions are created on demand: for each
different database backend, different subclasses are defined, to make
sure dispatch is routed to the right methods.

The reason for this is that other methods, unknown to this package,
might dispatch on the DBI class. One such example is *dbplyr* that
introduces specialized behaviors for many classes. The `explain()`
method calls the internal `db_explain()` method which uses `EXPLAIN
QUERY PLAN` for SQLite connections but `EXPLAIN` for unspecified
database connections. Without inheritance, *dbplyr* would use the
default method. This might lead to errors for other databases that do
not understand `EXPLAIN`.

``` r
iris_tbl %>%
  summarize_if(is.numeric, mean) %>% 
  explain()
#> Applying predicate on the first 100 rows
#> res4 <- dbSendQuery(conn1, structure("SELECT *\nFROM `iris`\nLIMIT 100", class = c("sql", 
#> "character")))
#> dbFetch(res4, n = 100)
#> ##   Sepal.Length Sepal.Width Petal.Length Petal.Width Species
#> ## 1          5.1         3.5          1.4         0.2  setosa
#> ## 2          4.9         3.0          1.4         0.2  setosa
#> ## 3          4.7         3.2          1.3         0.2  setosa
#> dbHasCompleted(res4)
#> ## [1] TRUE
#> dbClearResult(res4)
#> <SQL>
#> SELECT AVG(`Sepal.Length`) AS `Sepal.Length`, AVG(`Sepal.Width`) AS `Sepal.Width`, AVG(`Petal.Length`) AS `Petal.Length`, AVG(`Petal.Width`) AS `Petal.Width`
#> FROM `iris`
#> 
#> <PLAN>
#> dbGetQuery(conn1, structure("EXPLAIN QUERY PLAN SELECT AVG(`Sepal.Length`) AS `Sepal.Length`, AVG(`Sepal.Width`) AS `Sepal.Width`, AVG(`Petal.Length`) AS `Petal.Length`, AVG(`Petal.Width`) AS `Petal.Width`\nFROM `iris`", class = c("sql", 
#> "character")))
#> ##   id parent notused          detail
#> ## 1  3      0       0 SCAN TABLE iris
#>   id parent notused          detail
#> 1  3      0       0 SCAN TABLE iris
```

-----

Please note that the ‘dblog’ project is released with a [Contributor
Code of Conduct](CODE_OF_CONDUCT.md). By contributing to this project,
you agree to abide by its terms.

---
output: github_document
---

<!-- README.md is generated from README.Rmd. Please edit that file -->

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.path = "man/figures/README-",
  out.width = "100%"
)
pkgload::load_all()
```

# dblog

<!-- badges: start -->
[![Lifecycle: experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://www.tidyverse.org/lifecycle/#experimental)
[![CRAN status](https://www.r-pkg.org/badges/version/dblog)](https://cran.r-project.org/package=dblog)
<!-- badges: end -->

The goal of dblog is to implement logging for arbitrary DBI backends, similarly to Perl's [DBI::Log](https://metacpan.org/pod/DBI::Log).
This is useful for troubleshooting and auditing codes that access a database.
The initial use case for this package is to help debugging DBItest tests.

## Installation

You can install the released version of dblog from [CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("dblog")
```

Install the development version from GitHub using

``` r
# install.packages("devtools")
devtools::install_github("r-dbi/dblog")
```


## Basic example

The `dblog` driver wraps arbitrary drivers:

```{r init}
library(dblog)
drv <- dblog(RSQLite::SQLite())
```


All calls to DBI methods are logged, by default to the console.

```{r console}
conn <- dbConnect(drv, file = ":memory:")
dbWriteTable(conn, "iris", iris[1:3, ])
data <- dbGetQuery(conn, "SELECT * FROM iris")
dbDisconnect(conn)

data
```

The log is runnable R code!
Run it in a fresh session to repeat the operations, step by step or in an otherwise controlled fashion.

dblog is smart about DBI objects created or returned, and will assign a new variable name to each new object.
Cleared results or closed connections are not removed automatically.

## Logging options

Logging can be redirected to a file, optionally all outputs may be logged as well.
For example, use a collecting logger to output all calls and results after the fact.


```{r collect}
collecting_logger <- make_collect_logger()

drv <- dblog(RSQLite::SQLite(), logger = collecting_logger)
conn <- dbConnect(drv, file = ":memory:")
dbWriteTable(conn, "iris", iris[1:3, ])
data <- dbGetQuery(conn, "SELECT * FROM iris")
dbDisconnect(conn)

collecting_logger$retrieve()

ev <- evaluate::evaluate(collecting_logger$retrieve())
cat(unlist(ev, use.names = FALSE), sep = "\n")
```

## Logging complex operations

The full power is demonstrated when running with code where the underlying _DBI_ operations are not obvious:

```{r dplyr}
library(dplyr)

drv <- dblog(RSQLite::SQLite())
conn <- dbConnect(drv, file = ":memory:")
dbWriteTable(conn, "iris", iris[1:3, ])

src <- dbplyr::src_dbi(conn)
iris_tbl <- tbl(src, "iris")
iris_tbl %>%
  summarize_if(is.numeric, mean)
```

## Inheritance hierarchy

Despite the common suggestion to [prefer composition over inheritance](https://en.wikipedia.org/wiki/Composition_over_inheritance), the new logging classes are implemented as subclasses of the actual DBI classes.
Moreover, the class definitions are created on demand: for each different database backend, different subclasses are defined, to make sure dispatch is routed to the right methods.

The reason for this is that other methods, unknown to this package, might dispatch on the DBI class.
One such example is _dbplyr_ that introduces specialized behaviors for many classes.
The `explain()` method calls the internal `db_explain()` method which uses `EXPLAIN QUERY PLAN` for SQLite connections but `EXPLAIN` for unspecified database connections.
Without inheritance, _dbplyr_ would use the default method.
This might lead to errors for other databases that do not understand `EXPLAIN`.

```{r dplyr-explain}
iris_tbl %>%
  summarize_if(is.numeric, mean) %>% 
  explain()
```

---

Please note that the 'dblog' project is released with a
[Contributor Code of Conduct](CODE_OF_CONDUCT.md).
By contributing to this project, you agree to abide by its terms.

#' @include Connection.R
NULL

setClass("dblogResult")

make_result_class <- function(base_class) {

  template_name <- "dblogResult"
  class_name <- paste0(template_name, "-", base_class)
  all_base_classes <- c(template_name, base_class)

  if (isClass(class_name)) {
    return(class_name)
  }

  where <- parent.frame()

  setClass <- function(...) {
    methods::setClass(..., where = where, package = .packageName)
  }

  setMethod <- function(...) {
    methods::setMethod(..., where = where)
  }

  class <- setClass(class_name,
    contains = all_base_classes, slots = list(res = base_class, log_call = "function"))

  setMethod(
    "show", class_name,
    function(object) {
      cat("<dblogResult>\n")
      show(object@res)
    })

  setMethod(
    "dbClearResult", class_name,
    function(res, ...) {
      res@log_call(dbClearResult(res@res, !!! enquos(...)))
    })

  setMethod(
    "dbFetch", class_name,
    function(res, n = -1, ...) {
      res@log_call(dbFetch(res@res, n = n, !!! enquos(...)))
    })

  setMethod(
    "dbHasCompleted", class_name,
    function(res, ...) {
      res@log_call(dbHasCompleted(res@res, !!! enquos(...)))
    })

  setMethod(
    "dbGetInfo", class_name,
    function(dbObj, ...) {
      dbObj@log_call(dbGetInfo(dbObj@res, !!! enquos(...)))
    })

  setMethod(
    "dbIsValid", class_name,
    function(dbObj, ...) {
      dbObj@log_call(dbIsValid(dbObj@res, !!! enquos(...)))
    })

  setMethod(
    "dbGetStatement", class_name,
    function(res, ...) {
      res@log_call(dbGetStatement(res@res, !!! enquos(...)))
    })

  setMethod(
    "dbColumnInfo", class_name,
    function(res, ...) {
      res@log_call(dbColumnInfo(res@res, !!! enquos(...)))
    })

  setMethod(
    "dbGetRowCount", class_name,
    function(res, ...) {
      res@log_call(dbGetRowCount(res@res, !!! enquos(...)))
    })

  setMethod(
    "dbGetRowsAffected", class_name,
    function(res, ...) {
      res@log_call(dbGetRowsAffected(res@res, !!! enquos(...)))
    })

  setMethod(
    "dbBind", class_name,
    function(res, params, ...) {
      res@log_call(dbBind(res@res, params, !!! enquos(...)))
      invisible(res)
    })

  class_name
}

make_log_call <- function(obj_name, log_obj) {
  force(obj_name)
  force(log_obj)

  s4_dict <- collections::Stack()

  find_s4_dict <- function(x) {
    all_s4 <- s4_dict$as_list()
    for (i in seq_along(all_s4)) {
      s4_i <- all_s4[[i]]
      if (identical(s4_i$obj, x)) {
        return(s4_i$name)
      }
    }

    # Not found
    x
  }

  add_s4_dict <- function(x) {
    if (inherits(x, "DBIResult")) prefix <- "res"
    else if (inherits(x, "DBIConnection")) prefix <- "conn"
    else if (inherits(x, "DBIDriver")) prefix <- "drv"
    else return(NULL)

    # Doesn't work yet (?)
    #if (!is.null(find_s4_dict(x))) return(NULL)

    all_s4 <- s4_dict$as_list()
    all_names <- purrr::map_chr(purrr::map(all_s4, "name"), as_string)

    prefix_names <- grep(paste0("^", prefix), all_names, value = TRUE)
    suffixes <- as.integer(gsub(paste0("^", prefix), "", prefix_names))

    max_prefix <- max(c(suffixes, 0L))

    new_name <- as.name(paste0(prefix, max_prefix + 1L))

    s4_dict$push(
      list(
        obj = x,
        name = new_name
      )
    )

    new_name
  }

  clear_s4_dict <- function() {
    s4_dict$clear()
  }


  log_call <- function(call) {
    quo <- enquo(call)
    expr <- quo_get_expr(quo)
    env <- quo_get_env(quo)

    args <- purrr::map(as.list(expr[-1]), ~ eval_tidy(., env = env))
    if (!is.null(obj_name)) {
      args[[1]] <- obj_name
    }

    args <- purrr::map(args, find_s4_dict)
    new_call <- call2(expr[[1]], !!!args)
    result <- NULL
    on.exit(log_obj$log(new_call, result))

    visible_quo <- rlang::new_quosure(call2(withVisible, expr), env)
    result <- eval_tidy(visible_quo)

    new_obj <- add_s4_dict(result$value)
    if (!is.null(new_obj)) {
      new_call <- call2("<-", new_obj, new_call)
      result$value <- wrap(result$value, new_obj, log_obj)
      result$visible <- FALSE
    }

    if (result$visible) {
      result$value
    } else {
      invisible(result$value)
    }
  }

  log_call
}

wrap <- function(x, name, log_obj) {
  if (inherits(x, "DBIDriver")) {
    class_name <- make_driver_class(class(x)[[1]])
    new(class_name, drv = x, log_call = make_log_call(name, log_obj))
  } else if (inherits(x, "DBIConnection")) {
    class_name <- make_connection_class(class(x)[[1]])
    new(class_name, conn = x, log_call = make_log_call(name, log_obj))
  } else if (inherits(x, "DBIResult")) {
    class_name <- make_result_class(class(x)[[1]])
    new(class_name, res = x, log_call = make_log_call(name, log_obj))
  } else {
    abort(paste0("Unknown class: ", paste(class(x), collapse = "/")))
  }
}

make_logger <- function(...) {
  logger <- list2(...)

  logger$log_call <- make_log_call(NULL, logger)
  logger
}

#' Logging parameters
#'
#' TBD.
#'
#' @export
get_default_logger <- function() {
  default_logger
}

format_console <- function(call, result, width = 80) {
  local_options(width = width)

  if (is.null(result)) {
    call <- call("try", call)
  }

  call_fmt <- safe_deparse(call, width = width)
  if (isTRUE(result$visible)) {
    output <- utils::capture.output(print(result$value))
    result_fmt <- paste0("## ", output)
  } else {
    result_fmt <- NULL
  }

  paste(c(call_fmt, result_fmt), collapse = "\n")
}

#' @export
#' @param path Passed on to [cat()] for the output.  Default: console output.
#' @rdname get_default_logger
make_text_logger <- function(path = NULL) {
  if (is.null(path)) {
    path <- ""
  }

  make_logger(
    log = function(call, result) {
      cat(format_console(call, result), "\n", sep = "", file = path, append = TRUE)
    }
  )
}

#' @export
#' @rdname get_default_logger
#' @param display `[logical(1)]`\cr
#'   Set to `TRUE` to display log entries in addition to writing
#'   to a file.
make_collect_logger <- function(display = FALSE) {
  queue <- collections::Queue()
  force(display)

  make_logger(
    log = function(call, result) {
      fmt <- format_console(call, result)

      if (isTRUE(display)) {
        writeLines(fmt)
      }

      queue$push(fmt)
    },

    retrieve = function() {
      glue::glue_collapse(as.character(queue$as_list()), sep = "\n")
    }
  )
}

default_logger <- make_text_logger()

#' @include Driver.R
NULL

setClass("dblogConnection")

#' @export
format.dblogConnection <- function(x, ...) {
  paste0("Logging<", format(x@conn), ">")
}

make_connection_class <- function(base_class) {

  template_name <- "dblogConnection"
  class_name <- paste0(template_name, "-", base_class)
  all_base_classes <- c(template_name, base_class)

  if (isClass(class_name)) {
    return(class_name)
  }

  where <- parent.frame()

  setClass <- function(...) {
    methods::setClass(..., where = where, package = .packageName)
  }

  setMethod <- function(...) {
    methods::setMethod(..., where = where)
  }

  class <- setClass(class_name,
    contains = all_base_classes, slots = list(conn = base_class, log_call = "function"))

  setMethod(
    "show", class_name,
    function(object) {
      cat("<dblogConnection>\n")
      show(object@conn)
    })

  setMethod(
    "dbIsValid", class_name,
    function(dbObj, ...) {
      dbObj@log_call(dbIsValid(dbObj@conn, !!! enquos(...)))
    })

  setMethod(
    "dbDisconnect", class_name,
    function(conn, ...) {
      conn@log_call(dbDisconnect(conn@conn, !!! enquos(...)))
    })

  setMethod(
    "dbSendQuery", c(class_name, "character"),
    function(conn, statement, ...) {
      conn@log_call(dbSendQuery(conn@conn, statement, !!! enquos(...)))
    })

  setMethod(
    "dbGetQuery", c(class_name, "character"),
    function(conn, statement, ...) {
      conn@log_call(dbGetQuery(conn@conn, statement, !!! enquos(...)))
    })

  setMethod(
    "dbSendStatement", c(class_name, "character"),
    function(conn, statement, ...) {
      conn@log_call(dbSendStatement(conn@conn, statement, !!! enquos(...)))
    })

  setMethod(
    "dbDataType", class_name,
    function(dbObj, obj, ...) {
      dbObj@log_call(dbDataType(dbObj@conn, obj, !!! enquos(...)))
    })

  setMethod(
    "dbQuoteString", c(class_name, "character"),
    function(conn, x, ...) {
      conn@log_call(dbQuoteString(conn@conn, x, !!! enquos(...)))
    })

  setMethod(
    "dbQuoteString", c(class_name, "SQL"),
    function(conn, x, ...) {
      conn@log_call(dbQuoteString(conn@conn, x, !!! enquos(...)))
    })

  setMethod(
    "dbQuoteIdentifier", class_name,
    function(conn, x, ...) {
      conn@log_call(dbQuoteIdentifier(conn@conn, x, !!! enquos(...)))
    })

  setMethod(
    "dbQuoteIdentifier", c(class_name, "character"),
    function(conn, x, ...) {
      conn@log_call(dbQuoteIdentifier(conn@conn, x, !!! enquos(...)))
    })

  setMethod(
    "dbQuoteIdentifier", c(class_name, "SQL"),
    function(conn, x, ...) {
      conn@log_call(dbQuoteIdentifier(conn@conn, x, !!! enquos(...)))
    })

  setMethod(
    "dbQuoteIdentifier", c(class_name, "Id"),
    function(conn, x, ...) {
      conn@log_call(dbQuoteIdentifier(conn@conn, x, !!! enquos(...)))
    })

  setMethod(
    "dbUnquoteIdentifier", class_name,
    function(conn, x, ...) {
      conn@log_call(dbUnquoteIdentifier(conn@conn, x, !!! enquos(...)))
    })

  setMethod(
    "dbQuoteLiteral", class_name,
    function(conn, x, ...) {
      conn@log_call(dbQuoteLiteral(conn@conn, x, !!! enquos(...)))
    })

  setMethod(
    "dbWriteTable", c(class_name, "character", "data.frame"),
    function(conn, name, value, overwrite = FALSE, append = FALSE, ...) {
      conn@log_call(dbWriteTable(conn@conn, name = name, value = value, overwrite = overwrite, append = append, !!! enquos(...)))
    })

  setMethod(
    "dbReadTable", c(class_name, "character"),
    function(conn, name, ...) {
      conn@log_call(dbReadTable(conn@conn, name = name, !!! enquos(...)))
    })

  setMethod(
    "dbListTables", class_name,
    function(conn, ...) {
      conn@log_call(dbListTables(conn@conn, !!! enquos(...)))
    })

  setMethod(
    "dbExistsTable", c(class_name, "character"),
    function(conn, name, ...) {
      conn@log_call(dbExistsTable(conn@conn, name, !!! enquos(...)))
    })

  setMethod(
    "dbListFields", c(class_name, "character"),
    function(conn, name, ...) {
      conn@log_call(dbListFields(conn@conn, name, !!! enquos(...)))
    })

  setMethod(
    "dbCreateTable", c(class_name),
    function(conn, name, fields, ...) {
      conn@log_call(dbCreateTable(conn@conn, name = name, fields = fields, !!! enquos(...)))
    })

  setMethod(
    "dbAppendTable", c(class_name),
    function(conn, name, value, ..., row.names = NULL) {
      conn@log_call(dbAppendTable(conn@conn, name = name, value = value, !!! enquos(...), row.names = row.names))
    })

  setMethod(
    "dbRemoveTable", c(class_name, "character"),
    function(conn, name, ...) {
      conn@log_call(dbRemoveTable(conn@conn, name, !!! enquos(...)))
    })

  setMethod(
    "dbGetInfo", class_name,
    function(dbObj, ...) {
      dbObj@log_call(dbGetInfo(dbObj@conn, !!! enquos(...)))
    })

  setMethod(
    "dbBegin", class_name,
    function(conn, ...) {
      conn@log_call(dbBegin(conn@conn, !!! enquos(...)))
    })

  setMethod(
    "dbCommit", class_name,
    function(conn, ...) {
      conn@log_call(dbCommit(conn@conn, !!! enquos(...)))
    })

  setMethod(
    "dbRollback", class_name,
    function(conn, ...) {
      conn@log_call(dbRollback(conn@conn, !!! enquos(...)))
    })

  class_name
}

#' DBI methods
#'
#' Implementations of pure virtual functions defined in the `DBI` package.
#' @name DBI
NULL

#' dblog driver and connector
#'
#' TBD.
#'
#' @export
#' @param drv Driver to be wrapped, object of class [DBI::DBIDriver-class].
#' @param logger Logger object, defaults to [get_default_logger()].
#' @import methods DBI
#' @examples
#' \dontrun{
#' #' library(DBI)
#' Rdblog::dblog()
#' }
dblog <- function(drv, logger = get_default_logger()) {
  if (is(drv, "duckdb_driver")) {
    # duckdb::duckdb()
    # Avoid suggesting package
    return(logger$log_call(!! new_call(call("::", sym("duckdb"), sym("duckdb")))))
  }
  expr <- gsub(", [)]$", ")", deparse(drv))
  quo <- parse(text = expr)[[1]]
  logger$log_call(!! quo)
}

#' @param cnr Connector to be wrapped, object of class [DBI::DBIConnector-class].
#' @rdname dblog
#' @export
dblog_cnr <- function(cnr, logger = get_default_logger()) {
  new(
    "DBIConnector",
    .drv = dblog(cnr@.drv, logger = logger),
    .conn_args = cnr@.conn_args
  )
}

setClass("dblogDriver")

make_driver_class <- function(base_class) {

  template_name <- "dblogDriver"
  class_name <- paste0(template_name, "-", base_class)
  all_base_classes <- c(template_name, base_class)

  if (isClass(class_name)) {
    return(class_name)
  }

  where <- parent.frame()

  setClass <- function(...) {
    methods::setClass(..., where = where, package = .packageName)
  }

  setMethod <- function(...) {
    methods::setMethod(..., where = where)
  }

  class <- setClass(class_name,
    contains = all_base_classes, slots = list(drv = base_class, log_call = "function"))

  setMethod(
    "show", class_name,
    function(object) {
      cat("<dblogDriver>\n")
      show(object@drv)
    })

  setMethod(
    "dbConnect", class_name,
    function(drv, ...) {
      drv@log_call(dbConnect(drv@drv, !!! enquos(...)))
    }
  )

  setMethod(
    "dbDataType", class_name,
    function(dbObj, obj, ...) {
      dbObj@log_call(dbDataType(dbObj@drv, obj, !!! enquos(...)))
    })

  setMethod(
    "dbIsValid", class_name,
    function(dbObj, ...) {
      dbObj@log_call(dbIsValid(dbObj@drv, !!! enquos(...)))
    })

  setMethod(
    "dbGetInfo", class_name,
    function(dbObj, ...) {
      dbObj@log_call(dbGetInfo(dbObj@drv, !!! enquos(...)))
    })

  class_name
}

test_that("output", {
  # Can't use expect_snapshot() because it uses deparse() itself

  expect_deparse(list(foo = "bar"))
  expect_deparse(list("\n" = 1))
  expect_deparse(list("\\n" = 1))
  expect_deparse(quote(list("\n" = 1)))
  expect_deparse(quote(list("\\n" = 1)))
  expect_deparse(rlang::set_names(list(1), ""))
})

safe_deparse <- function(x, width = getOption("width")) {
  out <- deparse(x, width.cutoff = width, backtick = TRUE)
  same <- tryCatch(
    identical(x, parse(text = out)[[1]]),
    error = function(e) FALSE
  )
  if (same) {
    return(glue::as_glue(out))
  }

  # Workaround for weird names like "" and "\n":
  glue::as_glue(deparse(
    x, width.cutoff = width, backtick = TRUE,
    control = c("keepNA", "keepInteger", "showAttributes")
  ))
}

expect_deparse <- function(x) {
  if (is.call(x)) {
    testthat::expect_identical(parse(text = safe_deparse(x))[[1]], x)
  } else {
    testthat::expect_identical(eval(parse(text = safe_deparse(x))[[1]]), x)
  }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Driver.R
\name{dblog}
\alias{dblog}
\alias{dblog_cnr}
\title{dblog driver and connector}
\usage{
dblog(drv, logger = get_default_logger())

dblog_cnr(cnr, logger = get_default_logger())
}
\arguments{
\item{drv}{Driver to be wrapped, object of class \link[DBI:DBIDriver-class]{DBI::DBIDriver}.}

\item{logger}{Logger object, defaults to \code{\link[=get_default_logger]{get_default_logger()}}.}

\item{cnr}{Connector to be wrapped, object of class \link[DBI:DBIConnector-class]{DBI::DBIConnector}.}
}
\description{
TBD.
}
\examples{
\dontrun{
#' library(DBI)
Rdblog::dblog()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make-log-call.R
\name{get_default_logger}
\alias{get_default_logger}
\alias{make_text_logger}
\alias{make_collect_logger}
\title{Logging parameters}
\usage{
get_default_logger()

make_text_logger(path = NULL)

make_collect_logger(display = FALSE)
}
\arguments{
\item{path}{Passed on to \code{\link[=cat]{cat()}} for the output.  Default: console output.}

\item{display}{\verb{[logical(1)]}\cr
Set to \code{TRUE} to display log entries in addition to writing
to a file.}
}
\description{
TBD.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Driver.R
\name{DBI}
\alias{DBI}
\title{DBI methods}
\description{
Implementations of pure virtual functions defined in the \code{DBI} package.
}
